# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import icalendar
import datetime
import pytz
from dateutil.rrule import rrulestr, rruleset, rrule, DAILY
import dateutil.parser
import sys
import x_wr_timezone
from collections import defaultdict
from icalendar.prop import vDDDTypes
from typing import Optional

if sys.version_info[0] == 2:
    # Python2 has no ZoneInfo. We can assume that pytz is used.
    _EPOCH = datetime.datetime.utcfromtimestamp(0)
    _EPOCH_TZINFO = pytz.UTC.localize(_EPOCH)
    def timestamp(dt):
        """Return the time stamp of a datetime"""
        # from https://stackoverflow.com/a/35337826
        if dt.tzinfo:
            return (dt - _EPOCH_TZINFO).total_seconds()
        return (dt - _EPOCH).total_seconds()
else:
    def timestamp(dt):
        """Return the time stamp of a datetime"""
        return dt.timestamp()

def convert_to_date(date):
    """converts a date or datetime to a date"""
    return datetime.date(date.year, date.month, date.day)

def convert_to_datetime(date, tzinfo):
    """converts a date to a datetime"""
    if isinstance(date, datetime.datetime):
        if date.tzinfo is None:
            if tzinfo is not None:
                if is_pytz(tzinfo):
                    return tzinfo.localize(date)
                return date.replace(tzinfo=tzinfo)
        elif tzinfo is None:
            return date.replace(tzinfo=None)
        return date
    elif isinstance(date, datetime.date):
        return datetime.datetime(date.year, date.month, date.day, tzinfo=tzinfo)
    return date

def time_span_contains_event(span_start, span_stop, event_start, event_stop, comparable=False):
    """Return whether an event should is included within a time span.

    - span_start and span_stop define the time span
    - event_start and event_stop define the event time
    - comparable indicates whether the dates can be compared.
        You can set it to True if you are sure you have timezones and
        date/datetime correctly or used make_comparable() before.

    Note that the stops are exlusive but the starts are inclusive.

    This is an essential function of the module. It should be tested in
    test/test_time_span_contains_event.py.
    """
    if not comparable:
        span_start, span_stop, event_start, event_stop = make_comparable((
            span_start, span_stop, event_start, event_stop
        ))
    assert event_start <= event_stop, "the event must start before it ends"
    assert span_start <= span_stop, "the time span must start before it ends"
    if event_start == event_stop:
        if span_start == span_stop:
            return event_start == span_start
        return span_start <= event_start < span_stop
    if span_start == span_stop:
        return event_start <= span_start < event_stop
    return event_start < span_stop and span_start < event_stop


def make_comparable(dates):
    """Make an list or tuple of dates comparable.

    Returns an list.
    """
    tzinfo = None
    for date in dates:
        tzinfo = getattr(date, "tzinfo", None)
        if tzinfo is not None:
            break
    return [convert_to_datetime(date, tzinfo) for date in dates]


def compare_greater(date1, date2):
    """Compare two dates if date1 > date2 and make them comparable before."""
    date1, date2 = make_comparable((date1, date2))
    return date1 > date2


def is_pytz(tzinfo):
    """Whether the time zone requires localize() and normalize().

    pytz requires these funtions to be used in order to correctly use the
    time zones after operations.
    """
    return hasattr(tzinfo , "localize")


class Repetition:
    """A repetition of an event."""

    ATTRIBUTES_TO_DELETE_ON_COPY = [
        "RRULE", "RDATE", "EXDATE"
    ]

    def __init__(self, source, start, stop, keep_recurrence_attributes=False, end_prop='DTEND'):
        """Create an event repetition.

        - source - the icalendar Event
        - start - the start date/datetime
        - stop - the end date/datetime
        - keep_recurrence_attributes - whether to copy or delete attributes
            mentioned in ATTRIBUTES_TO_DELETE_ON_COPY
        """
        self.source = source
        self.start = start
        self.stop = stop
        self.end_prop = end_prop
        self.keep_recurrence_attributes = keep_recurrence_attributes

    def as_vevent(self):
        """Create a shallow copy of the source event and modify some attributes."""
        revent = self.source.copy()
        revent["DTSTART"] = vDDDTypes(self.start)
        revent[self.end_prop] = vDDDTypes(self.stop)
        if not self.keep_recurrence_attributes:
            for attribute in self.ATTRIBUTES_TO_DELETE_ON_COPY:
                if attribute in revent:
                    del revent[attribute]
        for subcomponent in self.source.subcomponents:
            revent.add_component(subcomponent)
        return revent

    def is_in_span(self, span_start, span_stop):
        """Return whether the event is in the span."""
        return time_span_contains_event(span_start, span_stop, self.start, self.stop)

    def __repr__(self):
        """Debug representation with more info."""
        return "{}({{'UID':{}...}}, {}, {})".format(self.__class__.__name__, self.source.get("UID"), self.start, self.stop)

class RepeatedComponent:
    """Base class for RepeatedEvent, RepeatedJournal and RepeatedTodo"""

    def __init__(self, component, keep_recurrence_attributes=False):
        """Create an component which may have repetitions in it.

        - component - the icalendar component
        - keep_recurrence_attributes whether to copy or delete attributes in repetitions.
        """
        self.component = component
        self.start = self.original_start = self._get_component_start()
        self.end = self.original_end = self._get_component_end()
        self.keep_recurrence_attributes = keep_recurrence_attributes
        self.exdates = []
        self.exdates_utc = set()
        self.replace_ends = {}  # DTSTART -> DTEND # for periods
        exdates = component.get("EXDATE", [])
        for exdates in ((exdates,) if not isinstance(exdates, list) else exdates):
            for exdate in exdates.dts:
                self.exdates.append(exdate.dt)
                self.exdates_utc.add(self._unify_exdate(exdate.dt))
        self.rdates = []
        rdates = component.get("RDATE", [])
        for rdates in ((rdates,) if not isinstance(rdates, list) else rdates):
            for rdate in rdates.dts:
                if isinstance(rdate.dt, tuple):
                    # we have a period as rdate
                    self.rdates.append(rdate.dt[0])
                    self.replace_ends[timestamp(rdate.dt[0])] = rdate.dt[1]
                else:
                    # we have a date/datetime
                    self.rdates.append(rdate.dt)

        self.make_all_dates_comparable()

        self.duration = self.end - self.start
        self.rule = rule = rruleset(cache=True)
        _rule = component.get("RRULE", None)
        if _rule:
            # We don't support multiple RRULE yet, but we can support cases
            # where the same RRULE is erroneously repeated
            if isinstance(_rule, list):
                if len(_rule) > 0 and all(part == _rule[0] for part in _rule):
                    _rule = _rule[0]
                else:
                    raise ValueError("Don't yet support multiple distinct RRULE properties")
            self.rrule = self.create_rule_with_start(_rule.to_ical().decode())
            rule.rrule(self.rrule)
        else:
            self.rrule = None

        for exdate in self.exdates:
            rule.exdate(exdate)
        for rdate in self.rdates:
            rule.rdate(rdate)
        if not self.rrule or not self.rrule.until or not compare_greater(self.start, self.rrule.until):
            rule.rdate(self.start)

    def create_rule_with_start(self, rule_string):
        """Helper to create an rrule from a rule_string starting at the start of the component.

        Since the creation is a bit more complex, this function handles special cases.
        """
        try:
            return self.rrulestr(rule_string)
        except ValueError:
            # string: FREQ=WEEKLY;UNTIL=20191023;BYDAY=TH;WKST=SU
            # start: 2019-08-01 14:00:00+01:00
            # ValueError: RRULE UNTIL values must be specified in UTC when DTSTART is timezone-aware
            rule_list = rule_string.split(";UNTIL=")
            assert len(rule_list) == 2
            date_end_index = rule_list[1].find(";")
            if date_end_index == -1:
                date_end_index = len(rule_list[1])
            until_string = rule_list[1][:date_end_index]
            if self.is_all_dates:
                until_string = until_string[:8]
            elif self.tzinfo is None:
                # remove the Z from the time zone
                until_string = until_string[:-1]
            else:
                # we assume the start is timezone aware but the until value is not, see the comment above
                if len(until_string) == 8:
                    until_string += "T000000"
                assert len(until_string) == 15
                until_string += "Z" # https://stackoverflow.com/a/49991809
            new_rule_string = rule_list[0] + rule_list[1][date_end_index:] + ";UNTIL=" + until_string
            return self.rrulestr(new_rule_string)

    def rrulestr(self, rule_string):
        """Return an rrulestr with a start. This might fail."""
        rule = rrulestr(rule_string, dtstart=self.start, cache=True)
        rule.string = rule_string
        rule.until = until = self._get_rrule_until(rule)
        if is_pytz(self.start.tzinfo) and rule.until:
            # when starting in a time zone that is one hour off to the end,
            # we might miss the last occurrence
            # see issue 107 and test/test_issue_107_omitting_last_event.py
            rule = rule.replace(until=rule.until + datetime.timedelta(hours=1))
            rule.until = until
        return rule

    def _get_rrule_until(self, rrule):
        """Return the UNTIL datetime of the rrule or None if absent."""
        rule_list = rrule.string.split(";UNTIL=")
        if len(rule_list) == 1:
            return None
        assert len(rule_list) == 2, "There should be only one UNTIL."
        date_end_index = rule_list[1].find(";")
        if date_end_index == -1:
            date_end_index = len(rule_list[1])
        until_string = rule_list[1][:date_end_index]
        until = vDDDTypes.from_ical(until_string)
        return until

    def make_all_dates_comparable(self):
        """Make sure we can use all dates with eachother.

        Dates may be mixed and we have many of them.
        - date
        - datetime without timezone
        - datetime with timezone
        These three are not comparable but can be converted.
        """
        self.tzinfo = None
        dates = [self.start, self.end] + self.exdates + self.rdates
        self.is_all_dates = not any(isinstance(date, datetime.datetime) for date in dates)
        for date in dates:
            if isinstance(date, datetime.datetime) and date.tzinfo is not None:
                self.tzinfo = date.tzinfo
                break
        self.start = convert_to_datetime(self.start, self.tzinfo)

        self.end = convert_to_datetime(self.end, self.tzinfo)
        self.rdates = [convert_to_datetime(rdate, self.tzinfo) for rdate in self.rdates]
        self.exdates = [convert_to_datetime(exdate, self.tzinfo) for exdate in self.exdates]

    def _unify_exdate(self, dt):
        """Return a unique string for the datetime which is used to
        compare it with exdates.
        """
        if isinstance(dt, datetime.datetime):
            return timestamp(dt)
        return dt

    def within_days(self, span_start, span_stop):
        """Yield component Repetitions between the start (inclusive) and end (exclusive) of the time span."""
        # make dates comparable, rrule converts them to datetimes
        span_start = convert_to_datetime(span_start, self.tzinfo)
        span_stop = convert_to_datetime(span_stop, self.tzinfo)
        if compare_greater(span_start, self.start):
            # do not exclude an component if it spans across the time span
            span_start -= self.duration
        # NOTE: If in the following line, we get an error, datetime and date
        # may still be mixed because RDATE, EXDATE, start and rule.
        for start in self.rule.between(span_start, span_stop, inc=True):
            if isinstance(start, datetime.datetime) and is_pytz(start.tzinfo):
                # update the time zone in case of summer/winter time change
                start = start.tzinfo.localize(start.replace(tzinfo=None))
                # We could now well be out of bounce of the end of the UNTIL
                # value. This is tested by test/test_issue_20_exdate_ignored.py.
                if self.rrule is not None and self.rrule.until is not None and start > self.rrule.until and start not in self.rdates:
                    continue
            if self._unify_exdate(start) in self.exdates_utc or start.date() in self.exdates_utc:
                continue
            stop = self.replace_ends.get(timestamp(start), start + self.duration)
            yield Repetition(
                self.component,
                self.convert_to_original_type(start),
                self.convert_to_original_type(stop),
                self.keep_recurrence_attributes,
                self.end_prop
            )

    def convert_to_original_type(self, date):
        """Convert a date back if this is possible.

        Dates may get converted to datetimes to make calculations possible.
        This reverts the process where possible so that Repetitions end
        up with the type (date/datetime) that was specified in the icalendar
        component.
        """
        if not isinstance(self.original_start, datetime.datetime) and \
                not isinstance(self.original_end, datetime.datetime):
            return convert_to_date(date)
        return date

    def is_recurrence(self):
        """Whether this is a recurrence/modification of an event."""
        return "RECURRENCE-ID" in self.component

    def as_single_event(self) -> Optional[Repetition]:
        """Return as a singe event with no recurrences."""
        for repetition in self.within_days(self.start, self.start):
            return repetition


class RepeatedEvent(RepeatedComponent):
    """An event with repetitions created from an icalendar event."""
    end_prop = 'DTEND'

    def _get_component_start(self):
        """Return DTSTART"""
        ## Arguably, it may be considered a feature that this breaks if no DTSTART is set
        return self.component['DTSTART'].dt

    def _get_component_end(self):
        """Yield DTEND or calculate the end of the event based on DTSTART and DURATION."""
        ## an even may have DTEND or DURATION, but not both
        end = self.component.get("DTEND")
        if end is not None:
            return end.dt
        duration = self.component.get("DURATION")
        if duration is not None:
            return self.component["DTSTART"].dt + duration.dt
        return self.component["DTSTART"].dt

class RepeatedTodo(RepeatedComponent):
    end_prop = 'DUE'

    def _get_component_start(self):
        """Return DTSTART if it set, do not panic if it's not set."""
        ## easy case - DTSTART set
        start = self.component.get('DTSTART')
        if start is not None:
            return start.dt
        ## Tasks may have DUE set, but no DTSTART.
        ## Let's assume 0 duration and return the DUE
        due = self.component.get('DUE')
        if due is not None:
            return due.dt

        ## Assume infinite time span if neither is given
        ## (see the comments under _get_event_end)
        return datetime.date(*DATE_MIN)

    def _get_component_end(self):
        """Return DUE or DTSTART+DURATION or something"""
        ## Easy case - DUE is set
        end = self.component.get('DUE')
        if end is not None:
            return end.dt

        dtstart = self.component.get("DTSTART")

        ## DURATION can be specified instead of DUE.
        duration = self.component.get("DURATION")
        ## It is no requirement that DTSTART is set.
        ## Perhaps duration is a time estimate rather than an indirect
        ## way to set DUE.
        if duration is not None and dtstart is not None:
            return self.component["DTSTART"].dt + duration.dt

        ## According to the RFC, a VEVENT without an end/duration
        ## is to be considered to have zero duration.  Assuming the
        ## same applies to VTODO.
        if dtstart:
            return dtstart.dt

        ## The RFC says this about VTODO:
        ## > A "VTODO" calendar component without the "DTSTART" and "DUE" (or
        ## > "DURATION") properties specifies a to-do that will be associated
        ## > with each successive calendar date, until it is completed.
        ## It can be interpreted in different ways, though probably it may
        ## be considered equivalent with a DTSTART in the infinite past and DUE
        ## in the infinite future?
        return datetime.date(*DATE_MAX)

class RepeatedJournal(RepeatedComponent):
    end_prop = ''

    def _get_component_start(self):
        """Return DTSTART if it set, do not panic if it's not set."""
        ## according to the specification, DTSTART in a VJOURNAL is optional
        dtstart = self.component.get("DTSTART")
        if dtstart is not None:
            return dtstart.dt
        return datetime.date(*DATE_MIN)

    ## VJOURNAL cannot have a DTEND.  We should consider a VJOURNAL to
    ## describe one day if DTSTART is a date, and we can probably
    ## consider it to have zero duration if a timestamp is given.
    _get_component_end = _get_component_start


# The minimum value accepted as date (pytz + zoneinfo)
DATE_MIN = (1970, 1, 1)
# The maximum value accepted as date (pytz + zoneinfo)
DATE_MAX = (2038, 1, 1)

class UnsupportedComponent(ValueError):
    """This error is raised when a component is not supported, yet."""


class UnfoldableCalendar:
    '''A calendar that can unfold its events at a certain time.'''

    recurrence_calculators = {
        "VEVENT": RepeatedEvent,
        "VTODO": RepeatedTodo,
        "VJOURNAL": RepeatedJournal,
    }

    def __init__(self, calendar, keep_recurrence_attributes=False, components=["VEVENT"]):
        """Create an unfoldable calendar from a given calendar."""
        assert calendar.get("CALSCALE", "GREGORIAN") == "GREGORIAN", "Only Gregorian calendars are supported." # https://www.kanzaki.com/docs/ical/calscale.html
        self.repetitions = []
        for component_name in components:
            if component_name not in self.recurrence_calculators:
                raise UnsupportedComponent(f"\"{component_name}\" is an unknown name for a component. I only know these: {', '.join(self.recurrence_calculators)}.")
            for event in calendar.walk(component_name):
                recurrence_calculator = self.recurrence_calculators[component_name]
                self.repetitions.append(recurrence_calculator(event, keep_recurrence_attributes))

    @staticmethod
    def to_datetime(date):
        """Convert date inputs of various sorts into a datetime object."""
        if isinstance(date, int):
            date = (date,)
        if isinstance(date, tuple):
            date += (1,) * (3 - len(date))
            return datetime.datetime(*date)
        elif isinstance(date, str):
            # see https://docs.python.org/2/library/datetime.html#strftime-strptime-behavior
            if len(date) == 8:
                return datetime.datetime.strptime(date, "%Y%m%d")
            return datetime.datetime.strptime(date, "%Y%m%dT%H%M%SZ")
        return date

    def all(self):
        """Returns all events.

        I personally do not recommend to use this because
        this method is not documented and you may end up with lots of events most of which you may not use anyway."""
        # MAX and MIN values may change in the future
        return self.between(DATE_MIN, DATE_MAX)

    _DELTAS = [
        datetime.timedelta(days=1),
        datetime.timedelta(hours=1),
        datetime.timedelta(minutes=1),
        datetime.timedelta(seconds=1)
    ]

    def at(self, date):
        """Return all events within the next 24 hours of starting at the given day.

        - date can be a year like (2019,) or 2099
        - a month like (2019, 1) for January of 2019
        - a day like (2019, 1, 1) for the first of January 2019
        - a day with hours, (2019, 1, 1, 1)
        - a day with minutes, (2019, 1, 1, 1, 1)
        - a day with seconds, (2019, 1, 1, 1, 1, 1)
        """
        if isinstance(date, int):
            date = (date,)
        if isinstance(date, str):
            assert len(date) == 8 and date.isdigit(), "format yyyymmdd expected"
            date = (int(date[:4], 10), int(date[4:6], 10), int(date[6:]))
        if isinstance(date, datetime.datetime):
            return self.between(date, date)
        if isinstance(date, datetime.date):
            return self.between(date, date + datetime.timedelta(days=1))
        if len(date) == 1:
            return self.between((date[0], 1, 1), (date[0] + 1, 1, 1))
        if len(date) == 2:
            year, month = date
            if month == 12:
                return self.between((year, 12, 1), (year + 1, 1, 1))
            return self.between((year, month, 1), (year, month + 1, 1))
        dt = self.to_datetime(date)
        return self.between(dt, dt + self._DELTAS[len(date) - 3])

    def between(self, start, stop):
        """Return events at a time between start (inclusive) and end (inclusive)"""
        span_start = self.to_datetime(start)
        span_stop = self.to_datetime(stop)
        events = []
        events_by_id = defaultdict(dict) # UID (str) : RECURRENCE-ID(datetime) : event (Event)
        default_uid = object()
        def add_event(event):
            """Add an event and check if it was edited."""
            same_events = events_by_id[event.get("UID", default_uid)] # TODO: test what comes first
            recurrence_id = event.get("RECURRENCE-ID", event["DTSTART"]).dt # TODO: this is still wrong: what if there are different events at the same time?
            if isinstance(recurrence_id, datetime.datetime):
                recurrence_id = recurrence_id.date()
            other = same_events.get(recurrence_id, None)
            if other: # TODO: test that this is independet of order
                event_recurrence_id = event.get("RECURRENCE-ID", None)
                other_recurrence_id = other.get("RECURRENCE-ID", None)
                if event_recurrence_id is not None and other_recurrence_id is None:
                    events.remove(other)
                elif event_recurrence_id is None and other_recurrence_id is not None:
                    return
                else:
                    event_sequence = event.get("SEQUENCE", None)
                    other_sequence = other.get("SEQUENCE", None)
                    if event_sequence is not None and other_sequence is not None:
                        if event["SEQUENCE"] < other["SEQUENCE"]:
                            return
                        events.remove(other)
            same_events[recurrence_id] = event
            events.append(event)

        span_start_day = span_start
        if isinstance(span_start_day, datetime.datetime):
            span_start_day = span_start_day.replace(hour=0,minute=0,second=0)
        span_stop_day = span_stop
        if isinstance(span_stop_day, datetime.datetime):
            span_stop_day = span_stop.replace(hour=23,minute=59,second=59)

        # repetitions must be considered because the may remove events from the time span
        # see https://github.com/niccokunzmann/python-recurring-ical-events/issues/62
        remove_because_not_in_span = []
        for event_repetitions in self.repetitions:
            if event_repetitions.is_recurrence():
                repetition = event_repetitions.as_single_event()
                if repetition is None:
                    continue
                vevent = repetition.as_vevent()
                add_event(vevent)
                if not repetition.is_in_span(span_start, span_stop):
                    remove_because_not_in_span.append(vevent)
                continue
            for repetition in event_repetitions.within_days(span_start_day, span_stop_day):
                if compare_greater(repetition.start, span_stop):
                    break
                if repetition.is_in_span(span_start, span_stop):
                    add_event(repetition.as_vevent())

        for vevent in remove_because_not_in_span:
            try:
                events.remove(vevent)
            except ValueError:
                pass

        return events

def of(a_calendar, keep_recurrence_attributes=False, components=["VEVENT"]):
    """Unfold recurring events of a_calendar

    - a_calendar is an icalendar VCALENDAR component or something like that.
    - keep_recurrence_attributes - whether to keep attributes that are only used to calculate the recurrence.
    - components is a list of component type names of which the recurrences should be returned.
    """
    a_calendar = x_wr_timezone.to_standard(a_calendar)
    return UnfoldableCalendar(a_calendar, keep_recurrence_attributes, components)



"""Test that some attributes of the calendar objects
are kept if required.

See Issue 23 https://github.com/niccokunzmann/python-recurring-ical-events/issues/23.
"""
import pytest
from recurring_ical_events import of
import datetime
from pytz import UTC

RRULE = b'FREQ=DAILY;UNTIL=20160320T030000Z'
RDATE = b'20150705T190000Z'
EXDATE = b'20150705T190000Z'

class Default:
    @staticmethod
    def to_ical(): return None

@pytest.mark.parametrize("keywords,RRULE,RDATE,EXDATE",[
    ({}, None, None, None),
    ({"keep_recurrence_attributes": False}, None, None, None),
    ({"keep_recurrence_attributes": True}, RRULE, RDATE, EXDATE),
])
def test_keep_recurrence_attributes_default(calendars, keywords, RRULE, RDATE, EXDATE):
    calendar = calendars.raw.rdate2
    rcalendar = of(calendar, **keywords)
    events = rcalendar.at(2014)
    assert events
    for event in events:
        assert event.get("RRULE", Default).to_ical() == RRULE
        assert event.get("RDATE", Default).to_ical() == RDATE
        assert event.get("EXDATE", Default).to_ical() == EXDATE



"""This file tests whether the time input is correctly converted.

Also see test_convert_inputs.py
"""
import pytest
from recurring_ical_events import UnfoldableCalendar
from datetime import datetime
from pytz import utc

@pytest.mark.parametrize("input,output", [
    ((2019, 1, 1), datetime(2019, 1, 1)),
    ((2000, 12, 2), datetime(2000, 12, 2)),
    ((2000, 12, 2, 4), datetime(2000, 12, 2, 4)),
    ((2000, 12, 2, 4, 44), datetime(2000, 12, 2, 4, 44)),
    ((2000, 12, 2, 4, 44, 55), datetime(2000, 12, 2, 4, 44, 55)),
    (datetime(2001, 3, 12, tzinfo=utc), datetime(2001, 3, 12, tzinfo=utc)),
    ("20140511T000000Z", datetime(2014, 5, 11)),
    ("20150521", datetime(2015, 5, 21)),
])
def test_conversion(input, output):
    assert UnfoldableCalendar.to_datetime(input) == output




def test_one_deleted_event(calendars):
    events = calendars.each_week_but_one_deleted.all()
    assert len(events) == 7

def test_one_deleted_event(calendars):
    events = calendars.each_week_but_two_deleted.all()
    assert len(events) == 6



"""These tests are for Issue 28
https://github.com/niccokunzmann/python-recurring-ical-events/issues/28

"""
import pytest

def test_expected_amount_of_events(calendars):
    events = calendars.issue_28_rrule_with_UTC_endinginZ.between((2020, 5, 25),(2020, 9, 5))
    assert len(events) == 15




"""These tests make sure that you can select which components should be returned.

By default, it should be events.
If a component is not supported, an error is raised.
"""
import pytest


@pytest.mark.parametrize("components,count,calendar,message", [
    (None, 0, "issue-97-simple-todo", "by default, only events are returned"),
    (None, 0, "issue-97-simple-journal", "by default, only events are returned"),
    ([], 0, "rdate", "no components, no result"),
    ([], 0, "issue-97-simple-todo", "no components, no result"),
    ([], 0, "issue-97-simple-journal", "no components, no result"),
    (["VEVENT"], 0, "issue-97-simple-todo", "no events in the calendar"),
    (["VEVENT"], 0, "issue-97-simple-journal", "no events in the calendar"),

    (["VJOURNAL"], 0, "issue-97-simple-todo", "no journal, just a todo"),
    (["VTODO"], 1, "issue-97-simple-todo", "one todo is found"),

    (["VTODO"], 0, "issue-97-simple-journal", "no todo, just a journal"),
    (["VJOURNAL"], 1, "issue-97-simple-journal", "one journal is found"),

    (["VTODO", "VEVENT"], 0, "issue-97-simple-journal", "no todo, just a journal"),
    (["VJOURNAL", "VEVENT"], 1, "issue-97-simple-journal", "one journal is found"),
    (["VJOURNAL", "VEVENT", "VTODO"], 1, "issue-97-simple-journal", "one journal is found"),
])
def test_components_and_their_count(calendars, components, count, calendar, message):
    calendars.components = components
    repeated_components = calendars[calendar].at(2022)
    print(repeated_components)
    assert len(repeated_components) == count, f"{message}: {components}, {calendar}"


@pytest.mark.parametrize("component", [
    "VALARM", # existing but not supported, yet
    "vevent", # misspelled
    "ALDHKSJHK", # does not exist
])
def test_unsupported_component_raises_error(component, calendars):
    """If a component is not recognized, we want to inform the user."""
    with pytest.raises(ValueError) as error:
        calendars.components = [component]
        calendars.rdate
    assert f"\"{component}\"" in str(error)



from pytz import timezone
import datetime
import pytest

berlin = timezone("Europe/Berlin")

@pytest.mark.parametrize("date,time", [
    ((2019,3,20), berlin.localize(datetime.datetime(2019, 3, 20, 19))), # winter time, UTC+1
    ((2019,4,24), berlin.localize(datetime.datetime(2019, 4, 24, 19))), # summer time UTC+2
])
def test_daylight_saving_events(calendars, date, time):
    """Test the event 7uartkcnhf0elbvs8md0itrf6c@google.com."""
    event = calendars.daylight_saving_time.at(date)[0]
    expected_time = calendars.consistent_tz(time)
    print(event["UID"])
    print(event["DTEND"].dt)
    assert event["DTSTART"].dt == expected_time



#py3
#
# This is the benchmark for rrule speed
# see https://github.com/niccokunzmann/python-recurring-ical-events/issues/42
#

import icalendar
import os
import sys

HERE = os.path.dirname(__file__) or "."
sys.path.append(os.path.join(HERE, ".."))

import recurring_ical_events

# read utf
text = []
with open(os.path.join(HERE, 'issue42.ics'), "rb") as fobj:
  text += fobj.readlines()
text_utf = []
for i in text:
  text_utf.append(i.decode("utf-8"))
ical_string = "".join(text_utf)


calendar = icalendar.Calendar.from_ical(ical_string)

rec_calendar = recurring_ical_events.of(calendar)

for day in range(1,29):
  print("day",day)
  start_date = (2011, 11, day)
  events = rec_calendar.at(start_date)



"""
Test the DURATION property.

Not all events have an end.
Some events define no explicit end and some a DURATION.
RFC: https://www.kanzaki.com/docs/ical/duration.html

"""
import pytest

@pytest.mark.parametrize("date,count", [
    # event 3 days
    ("20180110", 1), ("20180111", 1), ("20180112", 1),
    ("20180109", 0), ("20180114", 0),
    # event 3 hours
    ((2018, 1, 15, 10), 1), ((2018, 1, 15, 11), 1), ((2018, 1, 15, 12), 1), 
    ((2018, 1, 15,  9), 0), ((2018, 1, 15, 14), 0), 
    # event with no duration nor end
    ((2018, 1, 20), 1),
    ((2018, 1, 19), 0), ((2018, 1, 21), 0),
])
def test_events_expected(date, count, calendars):
    events = calendars.duration.at(date)
    assert len(events) == count

@pytest.mark.parametrize("date,summary,expected_hours", [
    ("20190318", "original event", 1),
    ("20190319", "edited duration", 3),
    ("20190320", "original event", 1),
])
def test_duration_is_edited(calendars, date, summary, expected_hours):
    """Test that the duration of an event can be edited."""
    events = calendars.duration_edited.at(date)
    assert len(events) == 1
    event = events[0]
    event_hours = (event["DTEND"].dt - event["DTSTART"].dt).total_seconds() / 3600
    assert summary == event["SUMMARY"], "we should have the correct event"
    assert event_hours == expected_hours, "the duration is only edited in the edited event"



"""
This tests that subcomponents are carried over to different events.
"""

def test_subcomponents_are_compied(calendars):
    event = calendars.subcomponents.all()[0]
    assert event.subcomponents

def test_there_are_no_subcomponents(calendars):
    event = calendars.Germany.all()[0]
    assert not event.subcomponents




"""Imputs of different time zones should make a difference in the output"""
import pytest
import datetime
import pytz

@pytest.mark.parametrize("date,hours,timezone,number_of_events,calendar_name", [
# DTSTART;TZID=Europe/Berlin:20190304T000000
# time zone offset 6:07:00 between Europe/Berlin and Asia/Ho_Chi_Minh
    ((2019,3,4), 24, "Europe/Berlin", 1, "three_events"),
    ((2019,3,4), 24, "EST", 0, "three_events"),
    ((2019,3,4), 24, 'Asia/Ho_Chi_Minh', 1, "three_events"),
    ((2019,3,4), 1, 'Asia/Ho_Chi_Minh', 0, "three_events"),
    ((2019,3,4), 6, 'Asia/Ho_Chi_Minh', 0, "three_events"),
    ((2019,3,4), 7, 'Asia/Ho_Chi_Minh', 1, "three_events"),
# events that have no time zone, New Year
    ((2019,1,1), 1, 'Europe/Berlin', 1, "Germany"),
    ((2019,1,1), 1, 'Asia/Ho_Chi_Minh', 1, "Germany"),
    ((2019,1,1), 1, 'EST', 1, "Germany"),
])
def test_include_events_if_the_time_zone_differs(calendars, date, hours, timezone, number_of_events, calendar_name):
    """When the time zone is different, events can be included or
    excluded because they are in another time zone.
    """
    tzinfo = pytz.timezone(timezone)
    start = tzinfo.localize(datetime.datetime(*date))
    stop = start + datetime.timedelta(hours=hours)
    events = calendars[calendar_name].between(start, stop)
    if events:
        start = events[0]["DTSTART"].dt
        if isinstance(start, datetime.datetime):
            offset = start - start.replace(tzinfo=tzinfo)
            print("time zone offset {} between {} and {}".format(offset, start.tzinfo, tzinfo))
    assert len(events) == number_of_events, "in calendar {} and {} in {}".format(calendar_name, date, timezone)



import icalendar
import recurring_ical_events
try:
    from urllib.request import urlopen # Python 3
except ImportError:
    from urllib import urlopen # Python 2

start_date = (2019, 3, 5)
end_date =   (2019, 4, 1)
url = "http://tinyurl.com/y24m3r8f"

ical_string = urlopen(url).read() # https://stackoverflow.com/a/645318
calendar = icalendar.Calendar.from_ical(ical_string)
events = recurring_ical_events.of(calendar).between(start_date, end_date)
for event in events:
    start = event["DTSTART"].dt
    duration = event["DTEND"].dt - event["DTSTART"].dt
    print("start {} duration {}".format(start, duration))
# start 2019-03-18 04:00:00+01:00 duration 1:00:00
# start 2019-03-20 04:00:00+01:00 duration 1:00:00
# start 2019-03-19 04:00:00+01:00 duration 1:00:00
# start 2019-03-07 02:00:00+01:00 duration 1:00:00
# start 2019-03-08 01:00:00+01:00 duration 2:00:00
# start 2019-03-09 03:00:00+01:00 duration 0:30:00
# start 2019-03-10 duration 1 day, 0:00:00



import pytest
import os
import icalendar
import sys
import time
try:
    import zoneinfo as _zoneinfo
except ImportError:
    try:
        import backports.zoneinfo as _zoneinfo
    except ImportError:
        _zoneinfo = None
from x_wr_timezone import CalendarWalker


HERE = os.path.dirname(__file__)
REPO = os.path.dirname(HERE)

sys.path.append(REPO)

from recurring_ical_events import of

CALENDARS_FOLDER = os.path.join(HERE, "calendars")
# set the default time zone
# see https://stackoverflow.com/questions/1301493/setting-timezone-in-python
time.tzset()

class ICSCalendars:
    """A collection of parsed ICS calendars
    
    components is the argument to pass to the of function"""

    Calendar = icalendar.Calendar
    components = None

    def get_calendar(self, content):
        """Return the calendar given the content."""
        return self.Calendar.from_ical(content)

    def __getitem__(self, name):
        return getattr(self, name.replace("-", "_"))

    @property
    def raw(self):
        return ICSCalendars()

    def consistent_tz(self, dt):
        """Make the datetime consistent with the time zones used in these calendars."""
        assert dt.tzinfo is None or "pytz" in dt.tzinfo.__class__.__module__, "We need pytz time zones for now."
        return dt
        
    def _of(self, calendar):
        """Return the calendar but also with selected components."""
        if self.components is None:
            return of(calendar)
        return of(calendar, components=self.components)

for calendar_file in os.listdir(CALENDARS_FOLDER):
    calendar_path = os.path.join(CALENDARS_FOLDER, calendar_file)
    name = os.path.splitext(calendar_file)[0]
    with open(calendar_path, "rb") as file:
        content = file.read()
    @property
    def get_calendar(self, content=content):
        return self.get_calendar(content)
    attribute_name = name.replace("-", "_")
    setattr(ICSCalendars, attribute_name, get_calendar)

class Calendars(ICSCalendars):
    """Collection of calendars from recurring_ical_events"""

    def get_calendar(self, content):
        return self._of(ICSCalendars.get_calendar(self, content))

class ReversedCalendars(ICSCalendars):
    """All test should run in reversed item order.

    RFC5545:
        This memo imposes no ordering of properties within an iCalendar object.
    """

    def get_calendar(self, content):
        """Calendar traversing events in reversed order."""
        calendar = ICSCalendars.get_calendar(self, content)
        _walk = calendar.walk
        def walk(*args, **kw):
            """Return properties in reversed order."""
            return reversed(_walk(*args, **kw))
        calendar.walk = walk
        return self._of(calendar)


class ZoneInfoConverter(CalendarWalker):
    """Visit a calendar and change all time zones to ZoneInfo"""

    def walk_value_datetime(self, dt):
         """Chnage timezone of datetime to zoneinfo"""
         py_tz = dt.tzinfo
         if py_tz is None:
             return dt
         name = py_tz.zone #, py_tz.tzname(dt)
         new_tz = _zoneinfo.ZoneInfo(name) # create a zoneinfo from pytz time zone
         return dt.astimezone(new_tz)

class ZoneInfoCalendars(ICSCalendars):
    """Collection of calendars using zoneinfo.ZoneInfo and not pytz."""

    def __init__(self):
        assert _zoneinfo is not None, "zoneinfo must exist to use these calendars"
        self.changer = ZoneInfoConverter()

    def get_calendar(self, content):
        calendar = ICSCalendars.get_calendar(self, content)
        zoneinfo_calendar = self.changer.walk(calendar)
#        if zoneinfo_calendar is calendar:
#            pytest.skip("ZoneInfo not in use. Already tested..")
        return self._of(zoneinfo_calendar)

    def consistent_tz(self, dt):
        """To make the time zones consistent with this one, convert them to zoneinfo."""
        return self.changer.walk_value_datetime(dt)


calendar_params = [Calendars, ReversedCalendars]
if _zoneinfo is not None:
    calendar_params.append(ZoneInfoCalendars)

# for parametrizing fixtures, see https://docs.pytest.org/en/latest/fixture.html#parametrizing-fixtures
@pytest.fixture(params=calendar_params, scope="module")
def calendars(request):
    return request.param()

@pytest.fixture()
def todo():
    """Skip a test because it needs to be written first."""
    pytest.skip("This test is not yet implemented.")


@pytest.fixture(scope='module')
def zoneinfo():
    """Return the zoneinfo module if present, otherwise skip the test.

    Uses backports.zoneinfo or zoneinfo.
    """
    if _zoneinfo is None:
        pytest.skip("zoneinfo module not given. Use pip install backports.zoneinfo to install it.")
    return _zoneinfo


@pytest.fixture(scope='module')
def ZoneInfo(zoneinfo):
    """Shortcut for zoneinfo.ZoneInfo."""
    return zoneinfo.ZoneInfo



'''
These are tests concerning issue 4
https://github.com/niccokunzmann/python-recurring-ical-events/issues/4

It seems the rrule until parameter includes the last date
https://dateutil.readthedocs.io/en/stable/rrule.html
'''
import datetime
import pytest
import pytz

chicago = pytz.timezone("America/Chicago")

@pytest.mark.parametrize("start_time,end_time,expected_count", [
    (chicago.localize(datetime.datetime(2020, 12, 11, 8)),chicago.localize(datetime.datetime(2020, 12, 11, 15)),4),
    (chicago.localize(datetime.datetime(2020, 12, 11, 9)),chicago.localize(datetime.datetime(2020, 12, 11, 15)),3),
    (chicago.localize(datetime.datetime(2020, 12, 11, 10)),chicago.localize(datetime.datetime(2020, 12, 11, 15)),3),
])
def test_between(calendars, start_time, end_time, expected_count):
    events = calendars.issue_48_dst.between(start_time, end_time)
    assert len(events) == expected_count, "{} events expected between {} and {}".format(expected_count, start_time, end_time)


'''Test that different inputs are understood

Also see test_time_arguments.py
'''
import pytest

@pytest.mark.parametrize("start,stop,event_count", [
  (2020, 2021, 366),
  ((2020,), (2021,), 366),
  ((2019,2), (2020,2), 334),
  ((2019,2,4), (2019,5,21), 78),
  ("20190204", "20190521", 78),
  ("20190204T000000Z", "20190521T235959Z", 79), # 78 = that of the day
])
def test_calendar_between_allows_tuple(calendars, start, stop, event_count):
    events = calendars.one_day_event_repeat_every_day.between(start, stop)
    assert len(events) == event_count
